#/*
#//  The Natto View
#//
#//  Copyright (c) 1995-1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
#//                           Keio University (Matsushita Lab.)
#//
#//  Please first read the full copyright statement in the file LICENSE.
#//  The code in this file may be used, copied and redistributed under the
#//  same terms as perl itself (see the files README, Copying and Artistic
#//  in the perl distribution).
#*/

package Ident;
require 5.000;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(rident);
$VERSION = 1.00;

=head1 NAME

rident - identity check, query the login id of remote user

=head1 SYNOPSIS

    use Net::Ident;

    $rid = rident(SOCK);
    ($rid, $hostname) = rident(SOCK, 10);

=head DESCRIPTION

The C<rident> function gets the login id of remote user connecting by the
specified socket. The second parameter is the timeout in seconds. If not
specified, it will default to 5 seconds.

=head1 WARNING

C<rident()> uses C<alarm()> to implement the timeout, so don't set another
alarm while you are using it.

=cut

use strict;
use Carp;
use Socket;
use Symbol qw(gensym);
use vars qw($TIMEOUT);

sub IDENT_PORT { 113 }

$TIMEOUT = 5;


sub rident
{
    my($sock, $timeout) = @_;
    $timeout = $TIMEOUT unless $timeout;
    unless ($sock) {
	croak 'usage: $rid = rident($sock [, $timeout])' unless wantarray;
	croak 'usage: ($rid, $hostname) = rident($sock [, $timeout])';
    }
    my($thatport, $thatent) = sockaddr_in(getpeername($sock));
    my($thisport, $thisent) = sockaddr_in(getsockname($sock));
    my $rid_thataddr = sockaddr_in(IDENT_PORT, $thatent); 
    my $rid_thisaddr = sockaddr_in(0, "\0\0\0\0");

    my $rid;
    undef $rid;
    my $rid_sock = gensym;
    local($SIG{'ALRM'}) = $timeout ? sub { die }: $SIG{'ALRM'} || 'DEFAULT';
    eval {
        alarm($timeout) if $timeout;
        socket($rid_sock, PF_INET, SOCK_STREAM, getprotobyname('tcp')) || die;
	bind($rid_sock, $rid_thisaddr) || die;
	connect($rid_sock, $rid_thataddr) || die;
	select((select($rid_sock), $| = 1)[$[]);
    
        print $rid_sock "$thatport,$thisport\r\n";
	my $res = <$rid_sock>;
	alarm(0) if ($timeout);
	my($rid_thatport, $rid_thisport);
	($rid_thatport, $rid_thisport, $rid) =
	    ($res =~ /^\s*(\S+)\s*,\s*(\S+)\s*:\s*USERID\s*:[^:]*:\s*([^\r\n]*)/);
#        ($thatport != $rid_thatport || $thisport != $rid_thisport) && die;
    };
    alarm(0) if $timeout;
    undef $rid_sock;
    if (wantarray) {
        ($rid, gethostbyaddr($rid_thataddr, PF_INET));
    } else {
        $rid;
    }
}

1;
