#/*
#//  The Natto View
#//
#//  Copyright (c) 1995-1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
#//                           Keio University (Matsushita Lab.)
#//
#//  Please first read the full copyright statement in the file LICENSE.
#//  The code in this file may be used, copied and redistributed under the
#//  same terms as perl itself (see the files README, Copying and Artistic
#//  in the perl distribution).
#*/

package KanjiCode;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(sjis2euc sjis2jis jis2euc any2euc guess_kanji_code
             NO_KANJI JIS SJIS EUC);

use strict;

sub NO_KANJI () { 0; }
sub JIS      () { 1; }
sub SJIS     () { 2; }
sub EUC      () { 3; }


sub sjis2euc
{
    1 while ($_[0] =~ s/([^\x80-\xfc])(([\x81-\x9f\xe0-\xef][\x40-\x7e\x80-\xfc])+)([^\x81-\x9f\xe0-\xef])/$1.&k_sjis2euc($2).$4/egx);
}


sub k_sjis2euc
{
    my $fmt = 'C' . length($_[0]);
    my @s = unpack($fmt, $_[0]);
    my($i, $c1, $c2);
    foreach $i (0 .. $#s / 2) {
	$c1 = $s[$i * 2];
	$c2 = $s[$i * 2 + 1];

        # SJIS to EUC
	$c1 <<= 1;
        if ($c1 > 0x9f * 2){ $c1 += 0x80 - 0xb1 * 2 + 1; }
	else               { $c1 += 0x80 - 0x71 * 2 + 1; }
	if    ($c2 > 0x9e) { $c2 += 0x80 - 0x7e; $c1++; }
	elsif ($c2 > 0x7e) { $c2 += 0x80 - 0x20; }
	else               { $c2 += 0x80 - 0x1f; }

        $s[$i * 2]     = $c1;
	$s[$i * 2 + 1] = $c2;
    }
    pack($fmt, @s);
}


sub sjis2jis
{
    1 while ($_[0] =~ s/([^\x80-\xfc])(([\x81-\x9f\xe0-\xef][\x40-\x7e\x80-\xfc])+)([^\x81-\x9f\xe0-\xef])/$1.&k_sjis2jis($2).$4/egx);
}


sub k_sjis2jis
{
    my $fmt = 'C' . length($_[0]);
    my @s = unpack($fmt, $_[0]);
    my($i, $c1, $c2);
    foreach $i (0 .. $#s / 2) {
	$c1 = $s[$i * 2];
	$c2 = $s[$i * 2 + 1];

        # SJIS to JIS
	$c1 <<= 1;
        if ($c1 > 0x9f * 2){ $c1 -= 0xb1 * 2 - 1; }
	else               { $c1 -= 0x71 * 2 - 1; }
	if ($c2 > 0x9e)    { $c2 -= 0x7e; $c1++; }
	elsif ($c2 > 0x7e) { $c2 -= 0x20; }
	else               { $c2 -= 0x1f; }
    }
    "\033\$B" . pack($fmt, @s) . "\033\$B";
}


sub jis2euc
{
    1 while ($_[0] =~ s/\033\$[\\\@B]([^\033]+)\033\([HBJ]/&k_jis2euc($1)/eg);
}


sub k_jis2euc
{
    my $fmt = 'C' . length($_[0]);
    my @s = unpack($fmt, $_[0]);
    my $i;
    foreach $i (0 .. $#s) {
	$s[$i] |= 0x80;
    }
    pack($fmt, @s);
}


sub any2euc
{
    my $code = guess_kanji_code($_[0]);
    if ($code == NO_KANJI || $code == EUC) { return; }
    if ($code == JIS)  { jis2euc($_[0]); return; }
    if ($code == SJIS) { sjis2euc($_[0]); return; }
}


sub guess_kanji_code
{
    my($jis, $sjis, $euc) = (0, 0, 0);
    $jis++  while ($_[0] =~ /\033\$[\\\@B][^\033]+\033\([HBJ]/g);
    $sjis++ while ($_[0] =~ /[^\x80-\xfc]([\x81-\x9f\xe0-\xef][\x40-\x7e\x80-\xfc])+[^\x81-\x9f\xe0-\xef]/g);
    $euc++  while ($_[0] =~ /[\x00-\x7f]([\xa1-\xfe][\xa1-\xfe])+[\x00-\x7f]/g);
    return JIS  if ($jis  > $sjis && $jis  > $euc);
    return SJIS if ($sjis > $jis  && $sjis > $euc  / 5);
    return EUC  if ($euc  > $jis  && $euc  > $sjis * 5);
    return NO_KANJI;
}

1;
__END__

undef $/;
$_ = <>;
any2euc($_);
print;
1;



