/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                      Keio University (Matsushita Lab.)
//                      All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "EventManagerHTX.h"

#include <algorithm>
#include <functional>


EventManagerHTX::EventManagerHTX(
    XtAppContext app)
{
    if (!impl_) impl_ = new EventManagerHTX_impl(app);
}


int
EventManagerHTX_impl::register_event(
    SOCKET           fd,
    HTRequest*       req,
    SockOps          ops,
    HTEventCallback* callback,
    HTPriority)
{
    DEBUGOUT("register " << fd << ':' << ops);

    XtInputMask input_mask =
	((ops & rmask) ? XtInputReadMask  : 0) |
	((ops & wmask) ? XtInputWriteMask : 0) |
	((ops & emask) ? XtInputExceptMask: 0);
    
    event_vector::iterator i =
	find_if(events_.begin(), events_.end(), fd_match(fd));
    if (i != events_.end()) {
	(*i)->sockops |= ops;
	XtRemoveInput((*i)->inputid);
    }
    else {
	EventHTX* ep = new EventHTX;
	ep->fd = fd;
	events_.push_back(ep);
	i = events_.end();
	i--;
	(*i)->sockops = ops;
    }
    (*i)->htrequest = req;
    (*i)->htcallback = callback;
    (*i)->inputid = XtAppAddInput(app_, fd, (XtPointer) input_mask,
				  (XtInputCallbackProc) xt_input, *i);
    return HT_OK;
}


int
EventManagerHTX_impl::unregister_event(
    SOCKET  fd,
    SockOps ops)
{
    event_vector::iterator i =
	find_if(events_.begin(), events_.end(), fd_match(fd));
    if (i == events_.end()) return HT_OK;
    ops = (*i)->sockops & ~ops;

    DEBUGOUT("unregister " << fd ':' << ops);

    XtInputMask input_mask =
	((ops & rmask) ? XtInputReadMask  : 0) |
	((ops & wmask) ? XtInputWriteMask : 0) |
	((ops & emask) ? XtInputExceptMask: 0);

    XtRemoveInput((*i)->inputid);
    if (input_mask) {
	(*i)->sockops = ops;
	(*i)->inputid = XtAppAddInput(app_, fd, (XtPointer) input_mask,
				      (XtInputCallbackProc) xt_input, *i);
    }
    else {
	events_.erase(i);
    }
    return HT_OK;
}
