/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                      Keio University (Matsushita Lab.)
//                      All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef EVENTMANAGERHTX_H
#define EVENTMANAGERHTX_H
#ifdef __GNUG__
#pragma interface
#endif

#include <X11/Intrinsic.h>
#include "EventManagerHT.h"
#include "ptrvector.h"


struct EventHTX
{
    SOCKET           fd;
    HTRequest*       htrequest;
    HTEventCallback* htcallback;
    SockOps          sockops;
    XtInputId        inputid;
};


class EventManagerHTX_impl: public EventManagerHT_impl
{
    friend class EventManagerHTX;
 protected:
    EventManagerHTX_impl(XtAppContext app): app_(app) {}
    
    enum mask_t {
	rmask = FD_READ  | FD_ACCEPT | FD_CLOSE,
	wmask = FD_WRITE | FD_CONNECT,
	emask = FD_OOB 
    };
	
    virtual int register_event(SOCKET, HTRequest*, SockOps,
			       HTEventCallback*, HTPriority);
    virtual int unregister_event(SOCKET, SockOps);

    // Xt callback interface
    static void xt_input(EventHTX* me, XtInputId*)
	{ me->htcallback(me->fd, me->htrequest, me->sockops); }

 private:
    typedef ptrvector<EventHTX*> event_vector;
    event_vector events_;
    XtAppContext app_;
};


class EventManagerHTX: public EventManagerHT
{
 public:
    EventManagerHTX(XtAppContext app);
};


struct fd_match
{
    int fd_;
    fd_match(int fd): fd_(fd) {} 
    bool operator()(EventHTX* p) { return p->fd == fd_; }
    bool operator()(void* p) { return ((EventHTX*) p)->fd == fd_; }
};

#endif /* EVENTMANAGETHTX_H */
