/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef NODEINFO_H
#define NODEINFO_H
#ifdef __GNUG__
#pragma interface
#define __SINST__  /* Why need ? */
#endif

#include <vector>
#include <string>
#include "www.h"


class NodeInfo
{
 public:
    NodeInfo(string t = "Untitled", string m = "");

    bool is_redirect() { return http_status == HT_TEMP_REDIRECT ||
				http_status == HT_PERM_REDIRECT; }

    string title;     // document title
    int    doi;
    float  x, y, z, r;
    float  eye, eyebrow, mouth;
    bool   scanned;   // 
    bool   disabled;  // 
    vector<string> mark_keys; // 

    // for temporary use
    bool   drawn;
    bool   visited;    // used by depth first search
    int    written;    // serial number in writing
    float  x_2d, y_2d; // 2d layout

    // http header info
    int    http_status;
    string content_type;
    int    content_length;
    time_t last_get;
    time_t last_head;
    time_t expires;
    time_t last_modified;
    URL    redirect;

    string content;
    int    n_images;
    int    n_j_words;
    int    n_e_words;
    int    n_j_letters;
    int    n_e_letters;
    void   fill_variables();

    bool   icon_changed;
    bool   label_changed;

    // GL specific
    unsigned icon_list;   // display list of node icon
    unsigned label_list;  // display list of node label
};


class LinkInfo
{
 public:
    LinkInfo(string f = "", string a = "", string m = "");

    string frag;
    string anchor;   // <A HREF="...">anchor</A>
    string method;
	
    bool   scanned;  //

    // for temporary use
    bool   drawn;
    bool   visited;  // used by depth first search
};


inline bool
operator==(
    const LinkInfo& x,
    const LinkInfo& y)
{
    return x.frag == y.frag && x.anchor == y.anchor;
}

#endif /* NODEINFO_H */
