/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "NodeMapperUU.h"

#include <math.h>
#include <ctype.h>
#include "Web.h"


NodeMapperUU::NodeMapperUU():
    NodeMapper("URL (1st letter)", "URL (2nd letter)")
{}

void
NodeMapperUU::layout(
    Web::node* vp)
{
    if (!vp) return;
    NodeInfo& info = vp->info();
    string    file = vp->key().last_word();

    info.eye = info.eyebrow = info.mouth = 0.5;
    info.x = info.y = 0;

    for (unsigned i = 0, j = 0; i < file.length(); i++) {
	int ch = toupper(file[i]);
	if (isalpha(ch) || isdigit(ch)) {
	    if (isdigit(ch)) ch = 'Z' + 1;
	    if (j % 2 == 0)
		info.x += (ch - 'A') * pow(26, - (int) j++ / 2);
	    else 
		info.y += (ch - 'A') * pow(26, - (int) j++ / 2);
	}
    }
    info.x /= 27;
    info.y /= 27;
    info.z  = z_step() * info.doi + log(vp->out().size() + 1) * 0.02;
    float r = def_radius() * log(vp->in().size() + vp->out().size() + 2);
    if (info.r != r) {
	info.r = r;
	info.icon_changed = true;
    }
}


bool
NodeMapperUU::x_label(
    int     nth,
    float&  pos,
    string& label)
{
    if (nth < 0 || 13 < nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    nth *= 2;
    pos = (float) nth / 27;
    if (nth == 26) label = "0-9";
    else           label = 'A' + nth;
    return true;
}


bool
NodeMapperUU::y_label(
    int     nth,
    float&  pos,
    string& label)
{
    if (nth < 0 || 13 < nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    nth *= 2;
    pos = (float) nth / 27;
    if (nth == 26) label = "0-9";
    else           label = 'A' + nth;
    return true;
}
