/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                      Keio University (Matsushita Lab.)
//                      All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "WebFrame.h"

#include <strstream>
#include "WebIO.h"
#include "WebView.h"


WebFrame::WebFrame(
    Web*    w,
    string  server,
    u_short port):
    web_(w), server_(server), port_(port),
    rootview_(NULL), last_update_(time(NULL))
{
    web()->attach(this);
}


void
WebFrame::update(
    WebNotify notify)
{
    if (notify.msg.length()) message(notify.msg);
    if (notify.priority <= WebNotify::message) return;

    time_t t = time(NULL);
    if (notify.priority <= WebNotify::ignorable && t - last_update_ < 1)
	return;
    last_update_ = t;

    rootview_->update();
    if (notify.priority >= WebNotify::sync) show_node_info();
}


void
WebFrame::focus(
    URL url)
{
    Web::node v(url, NodeInfo(url.last_word()));
    Web::node_iterator vi = web()->nodes().find(&v);
    focus(vi != web()->nodes().end() ? *vi: web()->insert(v));
}


void
WebFrame::focus(
    Web::node* vp)
{
    if (!vp) return;
    web()->active_node(vp);
    show_node_info();
    scan();
}


void
WebFrame::reload()
{
    Web::node* vp = web()->active_node();
    if (!vp) return;
    for (Web::link_reverse_iterator ei = vp->out().rbegin();
	 ei != vp->out().rend(); ei++) web()->erase(*ei);
    vp->info().scanned = false;
    scan();
}


void
WebFrame::scan(
    Web::node* vp)
{
    if (!vp) vp = web()->active_node();
    if (!vp) return;
    if (!vp->info().scanned) invoke_extender(vp, 2);
    update(WebNotify::ignorable);
}


void
WebFrame::show_node_info()
{
    if (!web()->active_node()) return;
    strstream msg;
    msg << "<<TITLE>> " << web()->active_node()->info().title
	<< "  <<URL>> " << web()->active_node()->key() << ends;
    message(msg.str());
    msg.freeze(false);
}


WebReader*
WebFrame::invoke_reader(
    istream* in)
{
    WebReader* reader = new WebReader(web(), in, WebIO::auto_free);
    reader->run();
    return reader;
}


WebWriter*
WebFrame::invoke_writer(
    ostream* out)
{
    active_view()->calc_2d_layout();
    WebWriter* writer = new WebWriter(web(), out, WebIO::auto_free);
    writer->run();
    return writer;
}

