/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "WebFrameX.h"

#include <strstream>
#include <fstream>
#include "glex.h"
#include "WebIOX.h"
#include "WebViewX.h"
#include "WebExtenderHT.h"
#include "NodeMapperUU.h"
#include "NodeMapperLU.h"
#ifdef PABLO
#include "NodeMapperGenX.h"
#endif
#include "Preference.h"

#define XTINTER_CLASS WebFrameX    

WebFrameX::WebFrameX(
    Web*    web,
    string  server,
    u_short port,
    Widget  shell):
    WebFrame(web, server, port),
    shell_(shell), browser_(mozilla), access_mode_(direct_http),
    anchorbox_popup_(NULL), anchorbox_form_(NULL), anchorbox_list_(NULL),
    anchorbox_btnbar_(NULL), anchorbox_prev_active_(NULL)
{
    int dbuf[] = { GLEX_DBLBUF, None };
    int sbuf[] = { GLEX_SGLBUF, None };
    Display* dpy = XtDisplay(shell);

    bool is_dbuf = true;
    XVisualInfo* vi = glXChooseVisual(dpy, DefaultScreen(dpy), dbuf);
    if (!vi) {
	vi = glXChooseVisual(dpy, DefaultScreen(dpy), sbuf);
	if (!vi)
	    XtAppError(app_context(), "No RGB visual with depth buffer.");
	is_dbuf = false;
    }
    Colormap cmap = 0;
    if (preference.install_cmap && vi->c_class != TrueColor) {
	cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen), 
			       vi->visual, AllocNone);
	XtVaSetValues(shell_,
		      XmNvisual,   vi->visual,
		      XmNdepth,    vi->depth,
		      XmNcolormap, cmap, NULL);
    }
    else {
	cmap = GetShareableColormap(dpy, vi);
    }	

    // form /////////////////////////////////////
    form_ = XmCreateForm(shell_, "form", NULL, 0);
    XtManageChild(form_);

    // menubar //////////////////////////////////
#define MENU_MACRO_CODE
#include "menubar.menu"
    XtVaSetValues(menubar_,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_NONE,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);

    // sitebar //////////////////////////////////
    sitebar_ = XmCreateForm(form_, "sitebar", NULL, 0);
    XtVaSetValues(sitebar_,
		  XmNtopAttachment,    XmATTACH_WIDGET,
		  XmNtopWidget,        menubar_,
		  XmNbottomAttachment, XmATTACH_NONE,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    XtManageChild(sitebar_);

    sitelabel_ = XmCreateLabel/*Gadget*/(sitebar_, "sitelabel", NULL, 0);
    XtVaSetValues(sitelabel_,
		  XmNtraversalOn,      False,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_NONE, NULL);
    XtManageChild(sitelabel_);

    sitebtn_ = XmCreatePushButton/*Gadget*/(sitebar_, "sitebtn", NULL, 0);
    XtVaSetValues(sitebtn_,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_NONE,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    // XtManageChild(sitebtn_); // after sitebox
    XTINTER_FAST_ADD(sitebtn_, XmNactivateCallback, sitebtn_activate);

    sitebox_ = XmCreateTextField(sitebar_, "sitebox", NULL, 0);
    XtVaSetValues(sitebox_,
		  XmNeditMode,         XmSINGLE_LINE_EDIT,
		  XmNvalue,            HOME_PAGE,
		  XmNresizeHeight,     True,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_WIDGET,
		  XmNleftWidget,       sitelabel_,
		  XmNrightAttachment,  XmATTACH_WIDGET,
		  XmNrightWidget,      sitebtn_, NULL);
    XtManageChild(sitebox_);
    XtManageChild(sitebtn_);
    XTINTER_FAST_ADD(sitebox_, XmNactivateCallback, sitebox_activate);
    
    // statbar //////////////////////////////////
    statbar_ = XmCreateForm(form_, "statbar", NULL, 0);
    XtVaSetValues(statbar_,
		  XmNtopAttachment,    XmATTACH_NONE,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    // XtManageChild(statbar_);  // after glframe

    // statbox_ = XmCreateTextField(statbar_, "statbox", NULL, 0);
    statbox_ = XmCreateText(statbar_, "statbox", NULL, 0);
    XtVaSetValues(statbox_,
		  XmNeditable,              False,
		  XmNtraversalOn,           False,
		  XmNcursorPositionVisible, False,
		  XmNvalue,                 "NattoView " NATTOVIEW,
		  XmNresizeHeight,          True,
		  XmNtopAttachment,         XmATTACH_FORM,
		  XmNbottomAttachment,      XmATTACH_FORM,
		  XmNleftAttachment,        XmATTACH_FORM,
		  XmNrightAttachment,       XmATTACH_FORM, NULL);
    XtManageChild(statbox_);

    glframe_ = XmCreateForm(form_, "glframe", NULL, 0);
    XtVaSetValues(glframe_,
		  // XmNsensitive,      True,
		  // XmNtraversalOn,    True,
		  // XmNnavigationType, XmTAB_GROUP,
		  XmNfractionBase,     10000,
		  XmNtopAttachment,    XmATTACH_WIDGET,
		  XmNtopWidget,        sitebar_,
		  XmNbottomAttachment, XmATTACH_WIDGET,
		  XmNbottomWidget,     statbar_,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    XtManageChild(glframe_);
    XtManageChild(statbar_);
    
    rootview(new WebViewX(this, glframe_,
			  cmap, preference.install_cmap, is_dbuf));
    
    filebox_ = XmCreateFileSelectionDialog(shell_, "filebox", NULL, 0);
    XtUnmanageChild(XmFileSelectionBoxGetChild(filebox_,
					       XmDIALOG_HELP_BUTTON));

    aboutbox_ = XmCreateMessageDialog(shell_, "aboutbox", NULL, 0);
    XtUnmanageChild(XmMessageBoxGetChild(aboutbox_,
					 XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(aboutbox_,
					 XmDIALOG_CANCEL_BUTTON));
    XmString str =
	XmStringCreateLtoR("The Natto View version " NATTOVIEW "\n"
			   "Copyright (c) 1995-1997\n"
			   "Hidekazu Shiozawa\n"
			   "<shiozawa@myo.inst.keio.ac.jp>,\n"
			   "Keio University (Matsushita Laboratory)\n"
			   "Bug reports are welcome!!",
			   XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(aboutbox_, XmNmessageString, str, NULL);
    XmStringFree(str);
    
    anchorbox_create();
}


void
WebFrameX::message(
    string msg)
{
    // XmTextFieldSetString(statbox_, msg.c_str());
    XmTextSetString(statbox_, msg.c_str());
}


void
WebFrameX::update(
    WebNotify notify)
{
    WebFrame::update(notify);
    XmUpdateDisplay(shell_);
}


void
WebFrameX::save()
{
    XtManageChild(filebox_);
    XTINTER_FAST_ADD(filebox_, XmNcancelCallback, save_cancel);
    XTINTER_FAST_ADD(filebox_, XmNokCallback,     save_ok);
}


void
WebFrameX::save_ok(
    Widget w,
    XmFileSelectionBoxCallbackStruct* fs)
{
    char* fname = NULL;
    XmStringGetLtoR(fs->value, XmSTRING_DEFAULT_CHARSET, &fname);
    ofstream* ofs = new ofstream(fname);
    if (ofs) invoke_writer(ofs);
    save_cancel(w, fs);
}


void
WebFrameX::save_cancel(
    Widget, XmFileSelectionBoxCallbackStruct*)
{
    XTINTER_FAST_REMOVE(filebox_, XmNcancelCallback, save_cancel);
    XTINTER_FAST_REMOVE(filebox_, XmNokCallback,     save_ok);
    XtUnmanageChild(filebox_);
}


void
WebFrameX::load()
{
    XtManageChild(filebox_);
    XTINTER_FAST_ADD(filebox_, XmNcancelCallback, load_cancel);
    XTINTER_FAST_ADD(filebox_, XmNokCallback,     load_ok);
}


void
WebFrameX::load_ok(
    Widget w,
    XmFileSelectionBoxCallbackStruct* fs)
{
    char* fname = NULL;
    XmStringGetLtoR(fs->value, XmSTRING_DEFAULT_CHARSET, &fname);
    ifstream* ifs = new ifstream(fname);
    if (ifs) invoke_reader(ifs);
    load_cancel(w, fs);
}


void
WebFrameX::load_cancel(
    Widget, XmFileSelectionBoxCallbackStruct*)
{
    XTINTER_FAST_REMOVE(filebox_, XmNcancelCallback, load_cancel);
    XTINTER_FAST_REMOVE(filebox_, XmNokCallback,     load_ok);
    XtUnmanageChild(filebox_);
}

void
WebFrameX::sitebox_activate(
    Widget, XtPointer)
{
    char* str = XmTextFieldGetString(sitebox_);
    focus(URL(str));
    XtFree(str);
}


void
WebFrameX::sitebtn_activate(
    Widget, XtPointer)
{
    browse();
}


void
WebFrameX::anchorbox_sitebtn_activate(
    Widget, XtPointer)
{
    browse();
}


void
WebFrameX::anchorbox_sitebox_activate(
    Widget, XtPointer)
{
    char* str = XmTextFieldGetString(anchorbox_sitebox_);
    focus(URL(str));
    XtFree(str);
}


void
WebFrameX::about()
{
    XtManageChild(aboutbox_);
}


void
WebFrameX::access_via_linkdb()
{
    access_mode_ = via_linkdb;
}


void
WebFrameX::access_direct_http()
{
    access_mode_ = direct_http;
}


void
WebFrameX::mapper_uu()
{
    NodeMapper* m = web()->mapper();
    web()->mapper(new NodeMapperUU);
    delete m;
    web()->layout_all();
}


void
WebFrameX::mapper_lu()
{
    NodeMapper* m = web()->mapper();
    web()->mapper(new NodeMapperLU);
    delete m;
    web()->layout_all();
}


void
WebFrameX::anchorbox_create()
{
    anchorbox_popup_ = 
	XtCreatePopupShell("anchorbox_popup", topLevelShellWidgetClass,
			   shell_, NULL, 0);
    XTINTER_FAST_ADD(anchorbox_popup_,
		     XmNdestroyCallback, anchorbox_destroy);
    anchorbox_form_ = XmCreateForm(anchorbox_popup_, "anchorbox_form", NULL,0);
    XtManageChild(anchorbox_form_);

    anchorbox_sitebar_ =
	XmCreateForm(anchorbox_form_, "anchorbox_sitebar", NULL,0);
    XtVaSetValues(anchorbox_sitebar_,
		  XmNorientation,      XmHORIZONTAL,
		  XmNadjustLast,       True,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_NONE,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    XtManageChild(anchorbox_sitebar_);

    anchorbox_sitelabel_ = XmCreateLabel(anchorbox_sitebar_,
					 "anchorbox_sitelabel", NULL, 0);
    XtVaSetValues(anchorbox_sitelabel_,
		  XmNtraversalOn,      False,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_NONE, NULL);
    XtManageChild(anchorbox_sitelabel_); 

    anchorbox_sitebtn_ = XmCreatePushButton(anchorbox_sitebar_,
					    "anchorbox_sitebtn", NULL, 0);
    XtVaSetValues(anchorbox_sitebtn_,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_NONE,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    // XtManageChild(anchorbox_sitebtn_);  // after sitebox
    XTINTER_FAST_ADD(anchorbox_sitebtn_,
		     XmNactivateCallback, anchorbox_sitebtn_activate);

    anchorbox_sitebox_ =
	XmCreateTextField(anchorbox_sitebar_, "anchorbox_sitebox", NULL, 0);
    XtVaSetValues(anchorbox_sitebox_,
		  XmNeditMode,         XmSINGLE_LINE_EDIT,
		  XmNvalue,            HOME_PAGE,
		  XmNresizeHeight,     True,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_WIDGET,
		  XmNleftWidget,       anchorbox_sitelabel_,
		  XmNrightAttachment,  XmATTACH_WIDGET,
		  XmNrightWidget,      anchorbox_sitebtn_, NULL);
    XtManageChild(anchorbox_sitebox_);
    XtManageChild(anchorbox_sitebtn_);
    XTINTER_FAST_ADD(anchorbox_sitebox_,
		     XmNactivateCallback, anchorbox_sitebox_activate);

    anchorbox_btnbar_ =
	XmCreateRowColumn(anchorbox_form_, "anchorbox_btnbar", NULL, 0);
    XtVaSetValues(anchorbox_btnbar_,
		  XmNorientation,      XmHORIZONTAL,
		  XmNtopAttachment,    XmATTACH_NONE,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    // XtManageChild(anchorbox_btnbar_);  // after anchorbox_list

    Widget w;
    w = XmCreatePushButton(anchorbox_btnbar_, "focus", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, anchorbox_focus);
    w = XmCreatePushButton(anchorbox_btnbar_, "scan", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, anchorbox_scan);
    w = XmCreatePushButton(anchorbox_btnbar_, "browse", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, anchorbox_browse);
    w = XmCreatePushButton(anchorbox_btnbar_, "tour", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, anchorbox_tour);
    w = XmCreatePushButton(anchorbox_btnbar_, "close", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, anchorbox_close);

    anchorbox_list_ =
	XmCreateScrolledList(anchorbox_form_, "anchorbox_list", NULL, 0);
    XtVaSetValues(anchorbox_list_, XmNselectionPolicy,  XmBROWSE_SELECT, NULL);
    XtVaSetValues(XtParent(anchorbox_list_),
		  XmNtopAttachment,    XmATTACH_WIDGET,
		  XmNtopWidget,        anchorbox_sitebar_,
		  XmNbottomAttachment, XmATTACH_WIDGET,
		  XmNbottomWidget,     anchorbox_btnbar_,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    XtManageChild(anchorbox_list_);
    XtManageChild(anchorbox_btnbar_);
    XTINTER_FAST_ADD(anchorbox_list_,
		     XmNdefaultActionCallback, anchorbox_action);
    XTINTER_FAST_ADD(anchorbox_list_,
		     XmNbrowseSelectionCallback, anchorbox_selection);
}


void
WebFrameX::anchorbox()
{
    XtPopup(anchorbox_popup_, XtGrabNone);
}


void
WebFrameX::anchorbox_action(
    Widget w, XmListCallbackStruct* ls)
{
    int item = ls->item_position - 1;
    focus(anchorbox_items_[item]->to());
    anchorbox_selection(w, ls);
}


void
WebFrameX::anchorbox_selection(
    Widget, XmListCallbackStruct* ls)
{
    int item = ls->item_position - 1;
    web()->active_link(anchorbox_items_[item]);
    scan(anchorbox_items_[item]->to());
}



void
WebFrameX::anchorbox_destroy(
    Widget, XtPointer)
{
    anchorbox_popup_ = anchorbox_form_ = anchorbox_list_ = NULL;
    anchorbox_create();
}



static inline int ListGetSelectedPosFirst(Widget w)
{
    int* poslist= new int[512];
    int  count = 0;
    XmListGetSelectedPos(w, &poslist, &count);
    return (count > 0)? poslist[0]: 0;
    delete[] poslist;
}


void
WebFrameX::anchorbox_focus(
    Widget, XtPointer)
{
    int pos = ListGetSelectedPosFirst(anchorbox_list_);
    if (pos) focus(anchorbox_items_[pos - 1]->to());
}


void
WebFrameX::anchorbox_scan(
    Widget, XtPointer)
{
    int pos = ListGetSelectedPosFirst(anchorbox_list_);

    for (int i = 0; i < (int) anchorbox_items_.size(); i++) {
	XmListDeselectAllItems(anchorbox_list_);
	web()->active_link(anchorbox_items_[i]);
	scan(anchorbox_items_[i]->to());
	XmListSelectPos(anchorbox_list_, i + 1, False);
    }

    XmListDeselectAllItems(anchorbox_list_);
    if (pos) XmListSelectPos(anchorbox_list_, pos, False);
}


void
WebFrameX::anchorbox_browse(
    Widget, XtPointer)
{
    int pos = ListGetSelectedPosFirst(anchorbox_list_);
    if (!pos) return;
    
    pos--;
    web()->active_link(anchorbox_items_[pos]);
    scan(anchorbox_items_[pos]->to());
    browse(anchorbox_items_[pos]->to());
}


void
WebFrameX::anchorbox_tour(
    Widget, XtPointer)
{
    if (anchorbox_items_.empty()) return;
    int pos = ListGetSelectedPosFirst(anchorbox_list_);

    int item = pos++;
    if (item <= 0 || anchorbox_items_.size() <= (unsigned) item) {
	item = 0;
	pos  = 1;
    }
    XmListDeselectAllItems(anchorbox_list_);
    web()->active_link(anchorbox_items_[item]);
    scan(anchorbox_items_[item]->to());
    browse(anchorbox_items_[item]->to());
    XmListSelectPos(anchorbox_list_, pos, False);
}


void
WebFrameX::anchorbox_close(
    Widget, XtPointer)
{
    XtPopdown(anchorbox_popup_);
}


void
WebFrameX::show_node_info()
{
    Web::node* vp = web()->active_node();
    if (!vp) return;
    strstream title;
    title << "NattoView - " << vp->info().title << ends;
    XtVaSetValues(shell_, XmNtitle, title.str(), NULL);
    XtVaSetValues(anchorbox_popup_, XmNtitle, title.str(), NULL);
    title.freeze(false);
    XmTextFieldSetString(sitebox_, vp->key().absolute().c_str());
    XmTextFieldSetString(anchorbox_sitebox_, vp->key().absolute().c_str());

    if (anchorbox_prev_active_ != vp) {
	XmListDeleteAllItems(anchorbox_list_);
	while (!anchorbox_items_.empty()) anchorbox_items_.pop_back();
    }
    for (Web::link_iterator ei = vp->out().begin();
	 ei != vp->out().end(); ei++) {
	strstream item;
	item << (*ei)->info().anchor << " ("
	    << (*ei)->to()->key().relative(vp->key());
	if ((*ei)->info().frag.length()) item << '#' << (*ei)->info().frag;
	item << ')' << ends;
	XmString xms = XmStringCreate(item.str(), XmSTRING_DEFAULT_CHARSET);
	if (!XmListItemExists(anchorbox_list_, xms)) {
	    XmListAddItem(anchorbox_list_, xms, 0);
	    anchorbox_items_.push_back(*ei);
	}
	XmStringFree(xms);
	item.freeze(false);
    }
    anchorbox_prev_active_ = vp;
}


void
WebFrameX::tourbox()
{
    ((WebViewX*) rootview())->tourbox();
}



WebExtender*
WebFrameX::invoke_extender(
    Web::node* vp,
    int        depth)
{
    if (access_mode_ == via_linkdb) {
	WebExtenderX* extender =
	    new WebExtenderX(web(), app_context(), vp, depth,
			     server(), port(), WebIO::activate);
	extender->run();
	return extender;
    }
    else {
	WebExtenderHT* extender = new WebExtenderHT(web(), vp, depth);
	extender->run();
	return extender;
    }
}


WebReader*
WebFrameX::invoke_reader(
    istream* in)
{
    ifstream* ifs = DCAST(ifstream*, in);
    if (ifs) {
	WebReader* reader =
	    new WebReaderX(web(), app_context(), ifs, WebIO::auto_free);
	reader->run();
	return reader;
    }
    else {
	return WebFrame::invoke_reader(in);
    }
}


// Netscape Navigator Remote Control
#include <setjmp.h>
extern "C" {
int mozilla_remote_commands(Display *dpy, Window window, char **commands);
jmp_buf mozilla_env;
}

// NCSA Mosaic CCI
#ifdef MOSAIC_CCI
#include <cci.h>
#include "libcci.h"
#endif

void
WebFrameX::mozilla_open(
    URL url)
{
    const char* commands[2];
    
    if (!setjmp(mozilla_env)) {
	strstream cmd;
	cmd << "openURL(" << url << ")" << ends;
	commands[0] = cmd.str();
	commands[1] = NULL;
	mozilla_remote_commands(XtDisplay(form_), 0, commands);
	cmd.freeze(false);
	return;
    }
    else {
	// Mozzila is not present in the same screen.
	message("Launching Netscape...");
	strstream cmd;
	cmd << "netscape " << url << " &" << ends;
	int ret = system(cmd.str());
	cmd.freeze(false);
	if (ret != -1 && ret != 127) return;
    }
    message("ERROR: Netscape not found.");
}
    

void
#ifdef MOSAIC_CCI
WebFrameX::mosaic_get(
    URL url)
{
    MCCIPort cciport = MCCIConnect("localhost", CCI_PORT, NULL, NULL);
    if (!MCCIIsConnected(cciport)) return;
    MCCIGet(cciport, url.absolute().c_str(),
	    MCCI_OUTPUT_CURRENT, MCCI_ABSOLUTE, NULL);
    MCCIDisconnect(cciport);
}
#else
WebFrameX::mosaic_get(URL) {}
#endif


void
WebFrameX::browse(
    Web::node* vp)
{
    if (!vp) vp = web()->active_node();
    switch (browser_) {
     case mozilla: mozilla_open(vp->key());  break;
     case mosaic:  mosaic_get  (vp->key());  break;
    }
}

