/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                      Keio University (Matsushita Lab.)
//                      All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef WEBIO_H
#define WEBIO_H
#ifdef __GNUG__
#pragma interface
#endif

#include <istream>
#include <ostream>
#include <string>
#include "Web.h"


//////////////////////////////////////////////////////////////////////////////
//  WebIO -- base class of I/O objects
//
class WebIO //: public Thread ?
{
 public:
    enum flags_t { activate = 001, auto_free = 002 };
    
    // constructor
    WebIO(Web* web, int flags = 0): web_(web), flags_(flags) {}

    // run as a thread
    virtual void run() = 0;
    
    // member access
    Web* web()         { return web_; }
    int  flags() const { return flags_; }

 protected:
    // Protected destructor (ARM 12.2c): What a beautiful idea!!
    virtual ~WebIO() {}

 private:
    Web* web_;   // related Web
    int  flags_; // flags bits
};


//////////////////////////////////////////////////////////////////////////////
//  WebReader
//
class WebReader: public WebIO
{
 public:
    WebReader(Web*, istream*, int = 0);
    virtual void run();

    Web::node* first_node() { return first_node_; }
    
 protected:
    virtual ~WebReader();
    bool     read_line();
    istream* in()  { return in_; }

 private:
    Web::node*       first_node_;  // starting point
    istream*         in_;
    Web::node_vector vvec_;
    float            ver_;
    string           type_;
};


//////////////////////////////////////////////////////////////////////////////
//  WebWriter
//
class WebWriter: public WebIO
{
 public:
    WebWriter(Web*, ostream*, int = 0);
    virtual void run();
    
 protected:
    ~WebWriter();
    
 private:
    ostream* out_;
    string   type_;
};


//////////////////////////////////////////////////////////////////////////////
//  WebExtender
//
class WebExtender: public WebIO
{
  public:
    WebExtender(Web* web, Web::node* first, int depth, int flags = 0);
    WebExtender(Web* web, URL url, int depth, int flags = 0);

    Web::node* first_node()  { return first_node_; }
    int        depth() const { return depth_; }
    
 protected:
    virtual ~WebExtender();
	
 private:
    Web::node* first_node_;  // starting point
    int        depth_;       // depth
};

#endif /* WEBIO_H */
