/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                      Keio University (Matsushita Lab.)
//                      All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef WEBIOX_H
#define WEBIOX_H
#ifdef __GNUG__
#pragma interface
#endif

#include <fstream>
#include <string>
#include <X11/Intrinsic.h>
#include "socketstream.h"
#include "WebIO.h"


//////////////////////////////////////////////////////////////////////////////
//  WebReaderX -- 
//
class WebReaderX: public WebReader
{
 public:
    // primitive functions
    WebReaderX(Web*, XtAppContext, ifstream*, int flags = 0);

    virtual void run();
    XtAppContext app_context() { return app_; }
    
 protected:
    void read_callback();
    
    // Xt callback interface
    static void xt_read_callback(WebReaderX* me, int*, XtInputId*)
	{ me->read_callback(); }
    
 private:
    XtAppContext app_; // application context
    XtInputId    id_;  // input ID
};


class WebExtenderX: public WebExtender
{
 public:
    WebExtenderX(Web* web, XtAppContext app, Web::node* first, int depth,
		 string server, u_short port, int flags = 0);
    WebExtenderX(Web* web, XtAppContext app, URL url, int depth,
		 string server, u_short port, int flags = 0);
    virtual void run();
    XtAppContext app_context() { return app_; }

 protected:
    // make connection and read responce status
    bool connect(string hostname, u_short port);
    void read_status();
    
    // Xt callback interface
    static void xt_read_status(WebExtenderX* me, int*, XtInputId*)
	{ me->read_status(); }
    
 private:
    class Reader: public WebReaderX
    {
     public:
	Reader(WebExtenderX* ex, Web* web, XtAppContext app, socketstream* in):
	    extenderx_(ex),
	    WebReaderX(web, app, (ifstream*) in, ex->flags() & ~auto_free) {}
     protected:
	virtual ~Reader() { delete extenderx_; }
     private:
	WebExtenderX* extenderx_;  // related extender
    };

    XtAppContext  app_;     // applicatoin context
    string        server_;  // host name of LinkDB
    u_short       port_;    // port# of LinkDB
    Reader*       readerx_; // invoking reader
    socketstream* in_;      // socket
    XtInputId     id_;      // input ID of in_
};

#endif /* WEBIOX_H */
