/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef WEBVIEWGL_H
#define WEBVIEWGL_H
#ifdef __GNUG__
#pragma interface
#endif

//////////////////////////////////////////////////////////////////////////////
//  WebViewGL.h -- OpenGL specific display routines
//

#include <GL/gl.h>
#include "WebView.h"


//////////////////////////////////////////////////////////////////////////////
//  WebViewGL --
//
class WebViewGL: public WebView
{
 public:
    // primitive functions
    WebViewGL(WebFrame*);
    WebViewGL(WebViewGL* viewgl, int x, int y, int width, int height);
    virtual ~WebViewGL() {}
    
    // picking objects
    virtual void pick(int x, int y, vector<int>& sysobjs,
		      Web::node_vector& nodes, Web::link_vector& links);
    virtual void viewable_nodes(Web::node_vector&);

    // OpenGL buffer operations
    virtual void make_current() {}
    virtual void swap_buffers() {}
    virtual void get_raster_pos(Web::node* v, float& x, float& y);

    struct Hit { GLuint z, name; };

 protected:
    // initialization of the GL state machine
    void init_graphics();
    
    // utility routines
    virtual void resize(int x, int y, int width, int height);
    void         projection_matrix();
    void         modelview_matrix();
    
    // string rasterlization
    virtual GLubyte* string_bitmap(string str, int& w, int& h, int& base) = 0;
    
    // subroutines of update()
    virtual void update_prefix();
    virtual void update_suffix();
    
    // subroutines of draw() (overrided)
    virtual void draw_prefix();
    virtual void draw_suffix();
    virtual void draw_nodes_prefix();
    virtual void draw_nodes_suffix();
    virtual void draw_a_node(Web::node* vp, int status);
    virtual void draw_links_prefix();
    virtual void draw_links_suffix();
    virtual void draw_a_link(Web::link* vp, int status);
    virtual void draw_board();
    virtual void draw_cursor();

    //
    virtual void draw_a_node_icon (Web::node* vp, int status);
    virtual void draw_a_node_label(Web::node* vp, int status);
    virtual void draw_a_link_line (Web::link* ep, int status);
    
    enum { SEL_BUF_SIZE = 2048 };

    enum name_t {
	node_bit   = (1 << 30),
	link_bit   = (1 << 29),
	sysobj_bit = (1 << 28),
	cursor     = (sysobj_bit | 1)
    };

 private:
    GLubyte* fill_bitmap_;         // for background of name plates 
    int      fill_bytes_;          // size of fill_bitmap_ in bytes
    Web::node_vector node_table_;  // GLname-onto-node table
    Web::link_vector link_table_;  // GLname-onto-link table
};


inline bool operator<(const WebViewGL::Hit& h1, const WebViewGL::Hit& h2)
    { return h1.z < h2.z; }

#endif /* WEBVIEWGL_H */
