/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#define __SINST__  /* Why need ? */
#endif

#include <Xm/XmAll.h>
#include "linkview.h"
#include "WebViewX.h"

#include <setjmp.h>
#include <strstream>
#include <X11/keysym.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include "copy_line.h"
#include "socketstream.h"
#include "glex.h"


WebViewX::WebViewX(
    WebFrameX* framex,
    Widget     glframe,
    Colormap   colormap,
    bool       install_cmap,
    bool       double_buffer):
    WebViewGL((WebFrame*) framex),
    glxcx_(NULL), vi_(NULL), cmap_(colormap), install_cmap_(install_cmap),
    cursor_menu_(NULL), pick_menu_(NULL), view_menu_(NULL),
    double_buffer_(double_buffer), font_set_(NULL),
    x_drag_start_(0), y_drag_start_(0), x_drag_end_(0), y_drag_end_(0),
    tourbox_popup_(NULL), tourbox_form_(NULL),
    tourbox_list_(NULL), tourbox_btnbar_(NULL), pending_(false)
{
    int w = 0, h = 0;
    XtVaGetValues(glframe, XmNwidth, &w, XmNheight, &h, NULL);
    h -= 20;
    
    htransscale_ = XmCreateScale(glframe, "htransscale", NULL, 0);
    XtVaSetValues(htransscale_,
		  XmNorientation,         XmHORIZONTAL,
		  XmNprocessingDirection, XmMAX_ON_RIGHT, 
		  XmNminimum,             -100,
		  XmNmaximum,              100,
		  XmNvalue,               (int) (trans().x * 100),
		  XmNbottomAttachment,    XmATTACH_FORM,
		  XmNtopAttachment,       XmATTACH_NONE,
		  XmNleftAttachment,      XmATTACH_FORM,
		  XmNrightAttachment,     XmATTACH_SELF,
		  XmNwidth,               min(w / 2, 150) , NULL);
    XtManageChild(htransscale_);
    XTINTER_FAST_ADD(htransscale_, XmNvalueChangedCallback,
		     htransscale_changed);
    XTINTER_FAST_ADD(htransscale_, XmNdragCallback,
		     htransscale_changed);

    hrotscale_ = XmCreateScale(glframe, "hrotscale", NULL, 0);
    XtVaSetValues(hrotscale_,
		  XmNorientation,         XmHORIZONTAL,
		  XmNprocessingDirection, XmMAX_ON_RIGHT, 
		  XmNminimum,             -180,
		  XmNmaximum,              180,
		  XmNvalue,               (int) rot().z,
		  XmNbottomAttachment,    XmATTACH_FORM,
		  XmNtopAttachment,       XmATTACH_NONE,
		  XmNleftAttachment,      XmATTACH_WIDGET,
		  XmNleftWidget,          htransscale_,
		  XmNrightAttachment,     XmATTACH_FORM, NULL);
    XtManageChild(hrotscale_);
    XTINTER_FAST_ADD(hrotscale_, XmNvalueChangedCallback, hrotscale_changed);
    XTINTER_FAST_ADD(hrotscale_, XmNdragCallback,         hrotscale_changed);

    vtransscale_ = XmCreateScale(glframe, "vtransscale", NULL, 0);
    XtVaSetValues(vtransscale_,
		  XmNorientation,         XmVERTICAL,
		  XmNprocessingDirection, XmMAX_ON_TOP, 
		  XmNminimum,             -100,
		  XmNmaximum,              100,
		  XmNvalue,               (int) (trans().y * 100),
		  XmNbottomAttachment,    XmATTACH_SELF,
		  XmNheight,              min(h / 2, 150),
		  XmNtopAttachment,       XmATTACH_FORM,
		  XmNleftAttachment,      XmATTACH_NONE,
		  XmNrightAttachment,     XmATTACH_FORM, NULL);
    XtManageChild(vtransscale_);
    XTINTER_FAST_ADD(vtransscale_, XmNvalueChangedCallback,
		     vtransscale_changed);
    XTINTER_FAST_ADD(vtransscale_, XmNdragCallback,
		     vtransscale_changed);

    vrotscale_ = XmCreateScale(glframe, "vrotscale", NULL, 0);
    XtVaSetValues(vrotscale_,
		  XmNorientation,         XmVERTICAL,
		  XmNprocessingDirection, XmMAX_ON_BOTTOM, 
		  XmNminimum,             -180,
		  XmNmaximum,                0,
		  XmNvalue,               (int) rot().x,
		  XmNbottomAttachment,    XmATTACH_WIDGET,
		  XmNbottomWidget,        hrotscale_,
		  XmNtopAttachment,       XmATTACH_WIDGET,
		  XmNtopWidget,           vtransscale_,
		  XmNleftAttachment,      XmATTACH_NONE,
		  XmNrightAttachment,     XmATTACH_FORM, NULL);
    XtManageChild(vrotscale_);
    XTINTER_FAST_ADD(vrotscale_, XmNvalueChangedCallback, vrotscale_changed);
    XTINTER_FAST_ADD(vrotscale_, XmNdragCallback,         vrotscale_changed);

    if (!init(glframe, "glxarea")) return;
    XtVaSetValues(glxarea_,
		  XmNbottomAttachment, XmATTACH_WIDGET,
		  XmNbottomWidget,     hrotscale_,
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_WIDGET,
		  XmNrightWidget,      vrotscale_, NULL);
    XtManageChild(glxarea_);
}


bool
WebViewX::init(
    Widget      glframe,
    const char* name)
{
    int dbuf[] = { GLEX_DBLBUF, None };
    int sbuf[] = { GLEX_SGLBUF, None };
    Display* dpy = XtDisplay(glframe);

    if (double_buffer_)
	vi_ = glXChooseVisual(dpy, DefaultScreen(dpy), dbuf);
    if (!double_buffer_ || !vi_) {
	vi_ = glXChooseVisual(dpy, DefaultScreen(dpy), sbuf);
	if (!vi_) return false;
	double_buffer_ = false;
    }

    glxarea_ =
	XtVaCreateManagedWidget(name,
				glwMDrawingAreaWidgetClass, glframe,
				GLwNvisualInfo,      vi_,
				GLwNdoublebuffer,    double_buffer_ ? True: False,
				GLwNinstallColormap, install_cmap_  ? True: False,
				XmNcolormap,         cmap_,
				XmNtraversalOn,      True,
				// XmNsensitive,        True,
				// XmNnavigationType,   XmTAB_GROUP,
				XmNuserData,       this, NULL);

    XTINTER_FAST_ADD(glxarea_, GLwNginitCallback,  ginit);
    XTINTER_FAST_ADD(glxarea_, GLwNexposeCallback, expose);
    XTINTER_FAST_ADD(glxarea_, GLwNresizeCallback, xresize);
    XTINTER_FAST_ADD(glxarea_, GLwNinputCallback,  input);

    char** missing_chars = NULL;
    int    missing_count = 0;
    char*  def_string    = NULL;
    font_set_ =
	XCreateFontSet(XtDisplay(glxarea_), // "-*-*-medium-r-normal--14-*-*-*-*-*-*-*"
		       "-adobe-times-medium-r-normal--12-120-75-75-p-64-iso8859-1"
#ifdef JAPANESE
		       ",-misc-fixed-medium-r-normal--14-130-75-75-c-140-jisx0208.1983-0"
		       ",-misc-fixed-medium-r-normal--14-130-75-75-c-70-jisx0201.1976-0"
#endif
		       , &missing_chars, &missing_count, &def_string);

#define MENU_MACRO_CODE
#include "cursor.menu"

#define MENU_MACRO_CODE
#include "view.menu"

    return true;
}


WebViewX::~WebViewX()
{
    Display* dpy = XtDisplay(glxarea_);
    glXDestroyContext(XtDisplay(glxarea_), glxcx_);
    XFreeFontSet(dpy, font_set_);
    XtDestroyWidget(glxarea_);
    XtDestroyWidget(htransscale_);
    XtDestroyWidget(vtransscale_);
    XtDestroyWidget(hrotscale_);
    XtDestroyWidget(vrotscale_);
}


void
WebViewX::update_suffix()
{
    WebViewGL::update_suffix();
    XmScaleSetValue(hrotscale_,   (int) rot().z);
    XmScaleSetValue(vrotscale_,   (int) rot().x);
    XmScaleSetValue(htransscale_, (int)(trans().x * 100));
    XmScaleSetValue(vtransscale_, (int)(trans().y * 100));
}


void
WebViewX::ginit(
    Widget,
    GLwDrawingAreaCallbackStruct* cs)
{
    XtVaGetValues(glxarea_, GLwNvisualInfo, &vi_, NULL);
    glxcx_ =
	glXCreateContext(XtDisplay(glxarea_), vi_,
			 parent() ? DCAST(WebViewX*, parent())->glxcx_: None,
			 GL_TRUE);
    if (!glxcx_) XtAppError(app_context(), "Can't create rendering context.");
    init_graphics();
    if (!parent()) {
	// root view
	resize(0, 0, cs->width, cs->height);
    }
    else {
	// subview
	Position x = 0, y = 0;
	XtVaGetValues(glxarea_, XmNx, &x, XmNy, &y, NULL);
	y = rootview()->height() - (y + cs->height);
	resize(x, max(y, (Position) 0), cs->width, cs->height);
    }
}


void
WebViewX::swap_buffers()
{
    // make_current();
    GLwDrawingAreaSwapBuffers(glxarea_);
    glXWaitGL();
}


void 
WebViewX::xresize(
    Widget,
    GLwDrawingAreaCallbackStruct* cs)
{
    int old_width  = width();
    int old_height = height(); 
    if (!parent()) {
	// root view
	resize(0, 0, cs->width, cs->height);
    }
    else { 
	// subview
	Position x = 0, y = 0;
	XtVaGetValues(glxarea_, XmNx, &x, XmNy, &y, NULL);
	y = rootview()->height() - (y + cs->height);
	resize(x, max(y, (Position) 0), cs->width, cs->height);
    }
    if (cs->width < old_width && cs->height < old_height) {
	// In this case, expose() is not called.
	post_redisplay();
    }
}


void 
WebViewX::expose(
    Widget,
    GLwDrawingAreaCallbackStruct*)
{
    post_redisplay();
}


void
WebViewX::post_redisplay()
{
    if (pending_) return;
    pending_ = true;
    XtAppAddWorkProc(XtWidgetToApplicationContext(glxarea_),
		     (XtWorkProc) xt_redisplay, (XtPointer) this);
}


void
WebViewX::redisplay()
{
    update();
    pending_ = false;
}


Boolean
WebViewX::xt_redisplay(
    WebViewX* me)
{
    me->redisplay();
    return True;
}


GLubyte*
WebViewX::string_bitmap(
    string str,
    int&   width,
    int&   height,
    int&   base_line)
{
    Display* dpy = XtDisplay(glxarea_);
    
    XRectangle bounding_box = { 0, 0, 0, 0 };
    XmbTextExtents(font_set_, str.c_str(), str.length(), NULL, &bounding_box);
    width     = max(bounding_box.width,  (unsigned short) 1);
    height    = max(bounding_box.height, (unsigned short) 1);
    base_line = height + bounding_box.y;
		   
    int bwidth = (width + 7) / 8;  // width in bytes
    GLubyte* bitmap = new GLubyte[bwidth * height];
    memset(bitmap, 0, bwidth * height);
    
    Window pixmap =
	XCreatePixmap(dpy, XtWindow(glxarea_), bwidth * 8, height, 1);
    XGCValues values;
    values.background = values.foreground = 0;
    GC gc = XCreateGC(dpy, pixmap, GCBackground | GCForeground, &values);
    XSetForeground(dpy, gc, 0);
    XFillRectangle(dpy, pixmap, gc, 0, 0, bwidth * 8, height);
    XSetForeground(dpy, gc, 1);

    XmbTextItem text;
    text.chars    = str.c_str();
    text.nchars   = str.length();
    text.delta    = 0;
    text.font_set = font_set_;
    XmbDrawText(dpy, pixmap, gc, 0, height - base_line, &text, 1);
    XImage* image =
	XGetImage(dpy, pixmap, 0, 0, bwidth * 8, height, 1, XYPixmap);

    for (int y = 0; y < height; y++) {
	for (int x = 0; x < width; x++) {
	    if (XGetPixel(image, x, height - 1 - y))
		bitmap[y * bwidth + x / 8] |= 0x80 >> x % 8;
	}
    }
    
    XFreeGC(dpy, gc);
    XFreePixmap(dpy, pixmap);
    XDestroyImage(image);
    return bitmap;
}
    

void
WebViewX::htransscale_changed(
    Widget,
    XmScaleCallbackStruct* cs)
{
    trans_left_right_abs(cs->value / 100.0);
}


void
WebViewX::vtransscale_changed(
    Widget,
    XmScaleCallbackStruct* cs)
{
    trans_up_down_abs(cs->value / 100.0);
}


void
WebViewX::hrotscale_changed(
    Widget,
    XmScaleCallbackStruct* cs)
{
    rot_left_right_abs(cs->value);
}


void
WebViewX::vrotscale_changed(
    Widget,
    XmScaleCallbackStruct* cs)
{
    rot_up_down_abs(cs->value);
}


void
WebViewX::input(
    Widget,
    GLwDrawingAreaCallbackStruct* cs)
{
    switch (cs->event->type) {
     case ButtonPress:   button_press((XButtonEvent*) cs->event);   break;
     case KeyRelease:    key_release((XKeyEvent*) cs->event);       break;
     case ButtonRelease: button_release((XButtonEvent*) cs->event); break;
#ifdef PABLO
     case MotionNotify:  motion_notify((XMotionEvent*) cs->event);  break;
#endif
    }
    return;
}


#if 0
void
WebViewX::structure_event(Widget, XtPointer udata, XEvent* event, Boolean*)
{
    if (event->type == UnmapNotify) {
	WebViewX* me = (WebViewX*) udata;
	int button = 0;
	XtVaGetValues(me->popup_menu_, XmNwhichButton, &button, NULL);
	XtVaSetValues(me->popup_menu_, XmNwhichButton,
		      (button == Button1) ? Button2: Button1, NULL);
	XtDestroyWidget(XtParent(me->popup_menu_));
	me->popup_menu_ = NULL;
    }
}
#endif


void
WebViewX::button_press(
    XButtonEvent* event)
{
    vector<int>      sysobjs;
    Web::node_vector nodes;
    Web::link_vector links;
    int x = event->x;
    int y = height() - event->y;

    if (pick_menu_) XtDestroyWidget(XtParent(pick_menu_));
    pick_menu_ = NULL;
    
    switch (event->button) {
     case Button1:
	event->x -= 10;
	event->y -= 25;
	event->x_root -= 10;
	event->y_root -= 25;

	// The viewport coordinate is same as the window.
	pick(x, y, sysobjs, nodes, links);
	
	if (!sysobjs.empty()) {
	    // XAllowEvents(XtDisplay(glxarea_), SyncPointer, CurrentTime);
	    // XtVaSetValues(cursor_menu_, XmNwhichButton, Button1, NULL);
	    XmMenuPosition(cursor_menu_, event);
	    XtManageChild(cursor_menu_);
	    break;
	}

	if (!nodes.empty()) {
	    Widget glframe = XtParent(glxarea_);
	    pick_menu_ = XmCreatePopupMenu(glframe, "pickMenu", NULL, 0);
	    XmMenuPosition(pick_menu_, event);
	    XtManageChild(XmCreateLabelGadget(pick_menu_, "title", NULL, 0));
	    XtManageChild(XmCreateSeparatorGadget(pick_menu_, "separator",
						  NULL, 0));
	    create_node_menu(pick_menu_, nodes);
	    if (!links.empty()) {
		Widget menu =
		    XmCreatePulldownMenu(glframe, "pickLinkMenu", NULL, 0);
		create_link_menu(menu, links);
                Widget btn =
		    XmCreateCascadeButtonGadget(pick_menu_, "linkBtn",NULL,0);
		XtVaSetValues(btn, XmNsubMenuId, menu, NULL);
		XtManageChild(btn);
	    }
	    // XAllowEvents(XtDisplay(glxarea_), SyncPointer, CurrentTime);
	    // XtVaSetValues(pick_menu_, XmNwhichButton, Button1, NULL);
	    XtManageChild(pick_menu_);
	    break;
	}
	else if (!links.empty()) {
	    Widget glframe = XtParent(glxarea_);
	    pick_menu_ = XmCreatePopupMenu(glframe, "pickLinkMenu", NULL, 0);
	    XmMenuPosition(pick_menu_, event);
	    XtManageChild(XmCreateLabelGadget(pick_menu_, "title", NULL, 0));
	    XtManageChild(XmCreateSeparatorGadget(pick_menu_, "separator",
						  NULL, 0));
	    create_link_menu(pick_menu_, links);
	    // XAllowEvents(XtDisplay(glxarea_), SyncPointer, CurrentTime);
	    // XtVaSetValues(pick_menu_, XmNwhichButton, Button1, NULL);
	    XtManageChild(pick_menu_);
	    break;
	}
	break;
	
     case Button2:
#ifdef PABLO
	x_drag_start_ = x_drag_end_ = event->x;
	y_drag_start_ = y_drag_end_ = event->y;
	rubberband();
#endif
	break;

     default:
	// XtVaSetValues(view_menu_, XmNwhichButton, Button3, NULL);
	XmMenuPosition(view_menu_, event);
	XtManageChild(view_menu_);
	break;
    }
#if 0
    if (popup_menu_)
	XtAddEventHandler(popup_menu_, StructureNotifyMask, False,
			  structure_event, (XtPointer) this);
#endif
}


void
WebViewX::button_release(
    XButtonEvent* event)
{
#ifdef PABLO
    int w, h;

    switch (event->button) {
     case Button2:
	rubberband();
	w = abs(x_drag_end_ - x_drag_start_);
	h = abs(y_drag_end_ - y_drag_start_);
	if (w < 100) w = 100;
	if (h < 100) h = 100;
	open_subview(xpos() + min(x_drag_start_, x_drag_end_),
		     ypos() + height() -  max(y_drag_start_, y_drag_end_),
		     w, h);
	break;
     default:
	break;
    }
#endif PABLO
}


void
WebViewX::create_node_menu(
    Widget            menu,
    Web::node_vector& nodes)
{
    int j = 0;
    for (Web::node_vector::iterator vi = nodes.begin();
	 vi != nodes.end() && j < 10; vi++, j++) {
	Web::node* vp = *vi;

	const URL& url = vp->key();
	Widget btn =
	    XmCreatePushButtonGadget(menu, url.absolute().c_str(), NULL, 0);
	strstream buf;
	buf << vp->info().title << " (";
	if (web()->active_node())
	    buf << url.relative(web()->active_node()->key());
	else
	    buf << url.absolute();
	buf << ')' << ends;
	XmString xms = XmStringCreate(buf.str(), XmSTRING_DEFAULT_CHARSET);
	buf.freeze(false);
	
	XtVaSetValues(btn,
		      XmNlabelString, xms,
		      XmNuserData,    vp, NULL);
	XmStringFree(xms);
	XTINTER_FAST_ADD(btn, XmNactivateCallback, node_menu_item);
	XtManageChild(btn);
    }
}


void
WebViewX::create_link_menu(
    Widget            menu,
    Web::link_vector& links)
{
    int j = 0;
    for (Web::link_vector::iterator ei = links.begin();
	 ei != links.end() && j < 10; ei++, j++) {
	Web::link* ep = *ei;
	const URL& url1 = ep->from()->key();
	const URL& url2 = ep->to()->key();
	string name = url1.absolute() + url2.absolute();
	
	Widget btn = XmCreatePushButtonGadget(menu, name.c_str(), NULL, 0);
	strstream buf;
	buf << ep->info().anchor << " (";
	if (web()->active_node())
	    buf << url1.relative(web()->active_node()->key());
	else
	    buf << url1.last_word();
	buf << " -> " << url2.relative(url1) << ')' << ends;
	XmString xms = XmStringCreate(buf.str(), XmSTRING_DEFAULT_CHARSET);
	buf.freeze(false);

	XtVaSetValues(btn,
		      XmNlabelString, xms,
		      XmNuserData,    ep, NULL);
	XmStringFree(xms);
	XTINTER_FAST_ADD(btn, XmNactivateCallback, link_menu_item);
	XtManageChild(btn);
    }
}


void 
WebViewX::node_menu_item(
    Widget btn,
    XtPointer)
{
    if (pick_menu_) XtDestroyWidget(XtParent(pick_menu_));
    pick_menu_ = NULL;

    Web::node* vp;
    XtVaGetValues(btn, XmNuserData, &vp, NULL);
    frame()->focus(vp);
}


void 
WebViewX::link_menu_item(
    Widget btn,
    XtPointer)
{
    if (pick_menu_) XtDestroyWidget(XtParent(pick_menu_));
    pick_menu_ = NULL;

    Web::link* ep;
    XtVaGetValues(btn, XmNuserData, &ep, NULL);
    frame()->focus(ep->to());
}


void
WebViewX::key_release(
    XKeyEvent* event)
{
    char   buf[4];
    KeySym keysym;
    XLookupString(event, buf, 1, &keysym, NULL);
    
    switch (keysym) {
     case XK_Escape:
     case XK_Q:  Quit();             break;  // -> accelerator
#ifdef PABLO
     case XK_C:  close();            break;  // -> accelerator
     case XK_O:  set_origin();       break;  // -> accelerator
     case XK_N:  tourbox();          break;  // -> accelerator
     case XK_F:  render_face_icon(); break;  // -> accelerator
#endif
     case XK_H:  home_view();        break;  // -> accelerator
     case XK_T:  top_view();         break;  // -> accelerator
     case XK_W:  render_wireframe(); break;  // -> accelerator
     case XK_S:  render_surface();   break;  // -> accelerator

     case XK_U:  lift_up();          break;  // -> accelerator
     case XK_D:  lift_down();        break;  // -> accelerator
     case XK_A:  lift_down_all();    break;  // -> accelerator
     case XK_B:  browse();           break;  // -> accelerator
     case XK_L:  scan();             break;  // -> accelerator

     case XK_w:  trans_up();         break;  // -> accelerator
     case XK_z:  trans_down();       break;  // -> accelerator
     case XK_a:  trans_left();       break;  // -> accelerator
     case XK_s:  trans_right();      break;  // -> accelerator

     case XK_r:  rot_up();           break;  // -> accelerator
     case XK_c:  rot_down();         break;  // -> accelerator
     case XK_d:  rot_left();         break;  // -> accelerator
     case XK_f:  rot_right();        break;  // -> accelerator

     case XK_x:  zoom_in();          break;  // -> accelerator
     case XK_e:  zoom_out();         break;  // -> accelerator
    }
}


void
WebViewX::tourbox_create()
{
    tourbox_popup_ =
	XtCreatePopupShell("tourbox_popup", topLevelShellWidgetClass,
			   XtParent(glxarea_), NULL, 0);
    XTINTER_FAST_ADD(tourbox_popup_, XmNdestroyCallback, tourbox_destroy);
    tourbox_form_ = XmCreateForm(tourbox_popup_, "tourbox_form", NULL, 0);
    XtManageChild(tourbox_form_);
    
    tourbox_btnbar_ =
	XmCreateRowColumn(tourbox_form_, "tourbox_btnbar", NULL, 0);
    XtVaSetValues(tourbox_btnbar_,
		  XmNorientation,      XmHORIZONTAL,
		  XmNtopAttachment,    XmATTACH_NONE,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    // XtManageChild(tourbox_btnbar_);  // after tourbox_list

    Widget w;
    w = XmCreatePushButton/*Gadget*/(tourbox_btnbar_, "scan", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, tourbox_scan);
    w = XmCreatePushButton/*Gadget*/(tourbox_btnbar_, "browse", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, tourbox_browse);
    w = XmCreatePushButton/*Gadget*/(tourbox_btnbar_, "tour", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, tourbox_tour);
    w = XmCreatePushButton/*Gadget*/(tourbox_btnbar_, "update", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, tourbox_update);
    w = XmCreatePushButton/*Gadget*/(tourbox_btnbar_, "close", NULL, 0);
    XtManageChild(w);
    XTINTER_FAST_ADD(w, XmNactivateCallback, tourbox_close);

    tourbox_list_ =
	XmCreateScrolledList(tourbox_form_, "tourbox_list", NULL, 0);
    XtVaSetValues(tourbox_list_, XmNselectionPolicy,  XmBROWSE_SELECT, NULL);
    XtVaSetValues(XtParent(tourbox_list_),
		  XmNtopAttachment,    XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_WIDGET,
		  XmNbottomWidget,     tourbox_btnbar_,
		  XmNleftAttachment,   XmATTACH_FORM,
		  XmNrightAttachment,  XmATTACH_FORM, NULL);
    XtManageChild(tourbox_list_);
    XtManageChild(tourbox_btnbar_);
    XTINTER_FAST_ADD(tourbox_list_, XmNdefaultActionCallback, tourbox_action);
    XTINTER_FAST_ADD(tourbox_list_,
		     XmNbrowseSelectionCallback, tourbox_selection);
}


void
WebViewX::tourbox()
{
    if (!tourbox_popup_) tourbox_create();
    tourbox_update(NULL, NULL);
    XtPopup(tourbox_popup_, XtGrabNone);
}


void
WebViewX::tourbox_update(
    Widget, XtPointer)
{
    while (!tourbox_items_.empty()) tourbox_items_.pop_back();
    viewable_nodes(tourbox_items_);
    tourbox_makelist(-1);
}


void
WebViewX::tourbox_makelist(
    int item)
{
    XmListDeleteAllItems(tourbox_list_);
    for (unsigned i = 0; i < tourbox_items_.size(); i++) {
	strstream buf;
	buf << tourbox_items_[i]->info().title
	     << " (" << tourbox_items_[i]->key() << ')' << ends;
	XmString xms = XmStringCreate(buf.str(), XmSTRING_DEFAULT_CHARSET);
	XmListAddItem(tourbox_list_, xms, 0);
	XmStringFree(xms);
	buf.freeze(false);
    }
    if (0 <= item && (unsigned) item < tourbox_items_.size())
	XmListSelectPos(tourbox_list_, item + 1, False);
}


void
WebViewX::tourbox_action(
    Widget,
    XmListCallbackStruct* ls)
{
    int item = ls->item_position - 1;
    frame()->focus(tourbox_items_[item]);
    browse();
    tourbox_makelist(item);
}


void
WebViewX::tourbox_selection(
    Widget,
    XmListCallbackStruct* ls)
{
    int item = ls->item_position - 1;
    frame()->focus(tourbox_items_[item]);
    tourbox_makelist(item);
}


void
WebViewX::tourbox_destroy(
    Widget, XtPointer)
{
    tourbox_popup_ = tourbox_form_ = tourbox_list_ = NULL;
}


static inline int ListGetSelectedPosFirst(Widget w)
{
    int* poslist= new int[512];
    int  count = 0;
    XmListGetSelectedPos(w, &poslist, &count);
    return (count > 0)? poslist[0]: 0;
    delete[] poslist;
}


void
WebViewX::tourbox_browse(
    Widget, XtPointer)
{
    browse();
    int item = ListGetSelectedPosFirst(tourbox_list_) - 1;
    tourbox_makelist(item);
}


void
WebViewX::tourbox_scan(
    Widget, XtPointer)
{
    int item = ListGetSelectedPosFirst(tourbox_list_) - 1;
    for (int i = 0; i < (int) tourbox_items_.size(); i++) {
	frame()->scan(tourbox_items_[i]);
	tourbox_makelist(i);
    }
    tourbox_makelist(item);
}


void
WebViewX::tourbox_tour(
    Widget, XtPointer)
{
    if (tourbox_items_.empty()) return;
    int item = ListGetSelectedPosFirst(tourbox_list_);
    
    if (item <= 0 || tourbox_items_.size() <= (unsigned) item) item = 0;
    XmListDeselectAllItems(tourbox_list_);
    frame()->focus(tourbox_items_[item]);
    browse();
    tourbox_makelist(item);
}


void
WebViewX::tourbox_close(
    Widget, XtPointer)
{
    XtPopdown(tourbox_popup_);
}
