/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef WEBVIEWX_H
#define WEBVIEWX_H
#ifdef __GNUG__
#pragma interface
#endif

//////////////////////////////////////////////////////////////////////////////
//  WebViewX.h -- X (and GL) specific display routines
//

#include <vector>
#include <algorithm>
#include <Xm/XmAll.h>
#include <GL/GLwMDrawA.h>
#include "WebViewGL.h"
#include "linkview.h"
#include "xtinter.h"

class WebFrameX;


//////////////////////////////////////////////////////////////////////////////
//  WebViewX --
//
class WebViewX: public WebViewGL
{
 public:
    // primitive functions
    WebViewX(WebFrameX* framex, Widget glframe,
	     Colormap cmap, bool install_cmap, bool double_buffer);
#ifdef PABLO
    WebViewX(WebViewX* viewx, Widget glframe,
	     const char* name, int x, int y, int width, int height);
#endif
    virtual ~WebViewX();

    // Xt and GLX interface
    XtAppContext app_context(){ return XtWidgetToApplicationContext(glxarea_);}
    Colormap     colormap()     { return cmap_; }
    GLXContext   glx_context()  { return glxcx_; }
    bool         double_buffer(){ return double_buffer_; }

    // buffer operations (overrided)
    virtual void make_current() { GLwDrawingAreaMakeCurrent(glxarea_, glxcx_);}
    virtual void swap_buffers();

    void tourbox();
    
 protected:
    bool init(Widget glframe, const char* name);
#ifdef PABLO    
    virtual WebView* do_open_subview(int x, int y, int width, int height);
#endif
    virtual void update_suffix();
    
    void create_node_menu(Widget, Web::node_vector&);
    void create_link_menu(Widget, Web::link_vector&);

    // string rasterlization
    virtual GLubyte* string_bitmap(string str, int& w, int& h, int& base);
    
    // window callbacks
#define XTINTER_CLASS WebViewX
    XTINTER_FAST(ginit,               GLwDrawingAreaCallbackStruct*)
    XTINTER_FAST(xresize,             GLwDrawingAreaCallbackStruct*)
    XTINTER_FAST(expose,              GLwDrawingAreaCallbackStruct*)
    XTINTER_FAST(input,               GLwDrawingAreaCallbackStruct*)
    XTINTER_FAST(htransscale_changed, XmScaleCallbackStruct*)
    XTINTER_FAST(vtransscale_changed, XmScaleCallbackStruct*)
    XTINTER_FAST(hrotscale_changed,   XmScaleCallbackStruct*)
    XTINTER_FAST(vrotscale_changed,   XmScaleCallbackStruct*)
    XTINTER_FAST(node_menu_item,     XtPointer)
    XTINTER_FAST(link_menu_item,     XtPointer)
#undef XTINTER_CLASS    

    void button_press(XButtonEvent*);
    void key_release(XKeyEvent*);
    void button_release(XButtonEvent*);
#ifdef PABLO    
    void motion_notify(XMotionEvent*);
    void rubberband();
#endif
    void tourbox_create();
    void tourbox_makelist(int);
#define XTINTER_CLASS WebViewX    
    XTINTER_FAST(tourbox_destroy,   XtPointer)
    XTINTER_FAST(tourbox_action,    XmListCallbackStruct*)
    XTINTER_FAST(tourbox_selection, XmListCallbackStruct*)
    XTINTER_FAST(tourbox_browse,    XtPointer)
    XTINTER_FAST(tourbox_scan,      XtPointer)
    XTINTER_FAST(tourbox_tour,      XtPointer)
    XTINTER_FAST(tourbox_update,    XtPointer)
    XTINTER_FAST(tourbox_close,     XtPointer)
#undef XTINTER_CLASS    

    void post_redisplay();
    void redisplay();
    static Boolean xt_redisplay(WebViewX*);

    // pop-up menus
    
#define MENU_MACRO_CLASS WebViewX    
#include "cursor.menu"

#define MENU_MACRO_CLASS WebViewX    
#include "view.menu"

    // when selecting invalid menu
    void under_construction() { message("Not yet implemented."); }
    
 private:
    GLXContext   glxcx_;         // GLX rendering context 
    XVisualInfo* vi_;
    Colormap     cmap_;          // colormap of glxarea_
    bool         install_cmap_;  // installing cmap ?
    Widget       glxarea_;       // OpenGL window
    Widget       htransscale_;
    Widget       vtransscale_;
    Widget       hrotscale_;
    Widget       vrotscale_;
    Widget       cursor_menu_;
    Widget       pick_menu_;
    Widget       view_menu_;
    bool         double_buffer_; // double-buffered?
    XFontSet     font_set_;
    XButtonEvent btn_event_;
    Position     x_drag_start_;
    Position     y_drag_start_;
    Position     x_drag_end_;
    Position     y_drag_end_;

    Widget tourbox_popup_;
    Widget tourbox_form_;
    Widget tourbox_list_;
    Widget tourbox_btnbar_;
    Web::node_vector tourbox_items_;

    bool pending_;
};

#endif  /* WEBVIEWX_H */
