/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef DFS_H
#define DFS_H
#ifdef __GNUG__
#pragma interface
#endif

//////////////////////////////////////////////////////////////////////////////
//  dfs.h -- depth first search for graph
//

#include "limits.h"
#include "graph.h"


enum dfs_return { dfs_next, dfs_found, dfs_visited };
    

template <class Vertex, class Finder, class EContainer>
Vertex*
dfs_forward_aux(
    Vertex* vp,
    Finder& finder,
    int     depth,
    int     max_depth,
    EContainer*)
{
    if (depth >= max_depth) return NULL;
    
    for (EContainer::iterator i = (EContainer::iterator) vp->out().begin();
	 i != (EContainer::iterator) vp->out().end(); i++) {
	Vertex* p = (*i)->to();
	switch (finder(p, *i, depth)) {
	 case dfs_next:
	    p = dfs_forward_aux(p, finder, depth + 1, max_depth,
				(EContainer*) NULL);
	    if (p) return p;
	    else   break;
	 case dfs_found:
	    return vp;
	 case dfs_visited:
	    break;
	}
    }
    return NULL;
}


template <class Vertex, class Finder>
inline Vertex*
dfs_forward(
    Vertex* vp,
    Finder& finder,
    int     max_depth = INT_MAX)
{
    switch (finder(vp, NULL, 0)) {
     case dfs_next:
	return dfs_forward_aux(vp, finder, 1, max_depth, &vp->in());
     case dfs_found:
	return vp;
     case dfs_visited:
	return NULL;
    }
    return NULL;
}

#endif /* DFS_H */
