/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

// This file contains the code based on GLUT 3.1
// Copyright (c) Mark J. Kilgard, 1994.
//
// and OpenGL samples by SGI
// (c) Copyright 1993, Silicon Graphics, Inc.


#include "linkview.h"
#include "glex.h"

#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <algorithm>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xmu/Xmu.h>
#include <X11/Xmu/StdCmap.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>


Colormap
GetShareableColormap(
    Display*     dpy,
    XVisualInfo* vi)
{
    Status status =
	XmuLookupStandardColormap(dpy, vi->screen, vi->visualid, vi->depth,
				  XA_RGB_DEFAULT_MAP, False, True);
    if (status == 1) {
	int n = 0;
	XStandardColormap* std_cmaps;
	status = XGetRGBColormaps(dpy, RootWindow(dpy, vi->screen),
				  &std_cmaps, &n, XA_RGB_DEFAULT_MAP);
	if (status == 1) {
	    for (int i = 0; i < n; i++) {
		if (std_cmaps[i].visualid == vi->visualid) {
		    Colormap cmap = std_cmaps[i].colormap;
		    XFree(std_cmaps);
		    return cmap;
		}
	    }
	}
    }
    return XCreateColormap(dpy, RootWindow(dpy, vi->screen), 
			   vi->visual, AllocNone);
}


void
GetUnpackingMode(
    PackingMode& pm)
{
    glGetIntegerv(GL_UNPACK_SWAP_BYTES,  &pm.swap_bytes);
    glGetIntegerv(GL_UNPACK_LSB_FIRST,   &pm.lsb_first);
    glGetIntegerv(GL_UNPACK_ROW_LENGTH,  &pm.row_length);
    glGetIntegerv(GL_UNPACK_SKIP_ROWS,   &pm.skip_rows);
    glGetIntegerv(GL_UNPACK_SKIP_PIXELS, &pm.skip_pixels);
    glGetIntegerv(GL_UNPACK_ALIGNMENT,   &pm.alignment);
}


void
SetUnpackingMode(
    PackingMode& pm)
{
    glPixelStorei(GL_UNPACK_SWAP_BYTES,  pm.swap_bytes);
    glPixelStorei(GL_UNPACK_LSB_FIRST,   pm.lsb_first);
    glPixelStorei(GL_UNPACK_ROW_LENGTH,  pm.row_length);
    glPixelStorei(GL_UNPACK_SKIP_ROWS,   pm.skip_rows);
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, pm.skip_pixels);
    glPixelStorei(GL_UNPACK_ALIGNMENT,   pm.alignment);
}


void
WireSphere(
    GLfloat radius,
    GLint   slices,
    GLint   stacks)
{
    GLUquadricObj* quad_obj = gluNewQuadric();
    gluQuadricDrawStyle(quad_obj, GLU_LINE);
    gluSphere(quad_obj, radius, slices, stacks);
    gluDeleteQuadric(quad_obj);
}


void
SolidSphere(
    GLfloat radius,
    GLint   slices,
    GLint   stacks)
{
    GLUquadricObj* quad_obj = gluNewQuadric();
    gluQuadricDrawStyle(quad_obj, GLU_FILL);
    gluQuadricNormals(quad_obj, GLU_SMOOTH);
    gluSphere(quad_obj, radius, slices, stacks);
    gluDeleteQuadric(quad_obj);
}


void
StrokeText(
    GLfloat x,
    GLfloat y,
    GLfloat z,
    GLfloat angle,
    GLfloat zoom,
    string  str,
    size_t  len)
{
    if (len == 0) len = str.length();

    glPushMatrix();
    glTranslatef(x, y, z);
    glRotatef(angle, 0.0, 0.0, 1.0);
    glScalef(zoom, zoom, zoom);
    len = min(str.length(), len);
    for (size_t i = 0; i < len; i++)
	StrokeCharacter(&StrokeRoman, str[i]);
    glPopMatrix();
}


//////////////////////////////////////////////////////////////////////////////
//  The following function is based on glut_stroke.c in GLUT 3.1
//
///* Copyright (c) Mark J. Kilgard, 1994. */
//
///* This program is freely distributable without licensing fees
//   and is provided without guarantee or warrantee expressed or
//   implied. This program is -not- in the public domain. */

void
StrokeCharacter(
    StrokeFontPtr fontinfo,
    int           c)
{
    StrokeCharPtr ch;
    if (c < 0 || c >= fontinfo->num_chars || !(ch = &fontinfo->ch[c]))
	return;
    
    int i = ch->num_strokes;
    StrokePtr stroke = ch->stroke;
    for (; i > 0; i--, stroke++) {
	glBegin(GL_LINE_STRIP);
	int j = stroke->num_coords;
	CoordPtr coord = stroke->coord;
	for (; j > 0; j--, coord++)
	    glVertex2f(coord->x, coord->y);
	glEnd();
    }
    glTranslatef(ch->right, 0.0, 0.0);
}

//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//  The following code is based on glut_shape.c in GLUT 3.1.
//
///* Copyright (c) Mark J. Kilgard, 1994. */
//
///* This program is freely distributable without licensing fees
//   and is provided without guarantee or warrantee expressed or
//   implied. This program is -not- in the public domain. */
//
//  Again, glut_shapes.c is based on Open GL samples by SGI.
//
/**
(c) Copyright 1993, Silicon Graphics, Inc.

ALL RIGHTS RESERVED

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies and that
both the copyright notice and this permission notice appear in
supporting documentation, and that the name of Silicon
Graphics, Inc. not be used in advertising or publicity
pertaining to distribution of the software without specific,
written prior permission.

THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU
"AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR
OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  IN NO
EVENT SHALL SILICON GRAPHICS, INC.  BE LIABLE TO YOU OR ANYONE
ELSE FOR ANY DIRECT, SPECIAL, INCIDENTAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER,
INCLUDING WITHOUT LIMITATION, LOSS OF PROFIT, LOSS OF USE,
SAVINGS OR REVENUE, OR THE CLAIMS OF THIRD PARTIES, WHETHER OR
NOT SILICON GRAPHICS, INC.  HAS BEEN ADVISED OF THE POSSIBILITY
OF SUCH LOSS, HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
ARISING OUT OF OR IN CONNECTION WITH THE POSSESSION, USE OR
PERFORMANCE OF THIS SOFTWARE.

US Government Users Restricted Rights

Use, duplication, or disclosure by the Government is subject to
restrictions set forth in FAR 52.227.19(c)(2) or subparagraph
(c)(1)(ii) of the Rights in Technical Data and Computer
Software clause at DFARS 252.227-7013 and/or in similar or
successor clauses in the FAR or the DOD or NASA FAR
Supplement.  Unpublished-- rights reserved under the copyright
laws of the United States.  Contractor/manufacturer is Silicon
Graphics, Inc., 2011 N.  Shoreline Blvd., Mountain View, CA
94039-7311.

OpenGL(TM) is a trademark of Silicon Graphics, Inc.
*/


static inline void
normalize(
    GLfloat v[3])
{
    GLfloat d = sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
    if (d == 0.0) v[0] = d = 1.0;
    d = 1 / d;
    v[0] *= d;
    v[1] *= d;
    v[2] *= d;
}


#define X .525731112119133606
#define Z .850650808352039932

static GLfloat vdata[12][3] = {
    { -X,  0,  Z }, {  X,  0,  Z }, { -X,  0, -Z }, {  X,  0, -Z },
    {  0,  Z,  X }, {  0,  Z, -X }, {  0, -Z,  X }, {  0, -Z, -X },
    {  Z,  X,  0 }, { -Z,  X,  0 }, {  Z, -X,  0 }, { -Z, -X,  0 },
};

static int tindices[20][3] = {
    {  0,  4,  1 }, {  0,  9,  4 }, {  9,  5,  4 }, {  4,  5,  8 },
    {  4,  8,  1 }, {  8, 10,  1 }, {  8,  3, 10 }, {  5,  3,  8 },
    {  5,  2,  3 }, {  2,  7,  3 }, {  7, 10,  3 }, {  7,  6, 10 },
    {  7, 11,  6 }, { 11,  0,  6 }, {  0,  1,  6 }, {  6,  1, 10 },
    {  9,  0, 11 }, {  9, 11,  2 }, {  9,  2,  5 }, {  7,  2, 11 },
};


static void
subdivide(
    GLfloat n1[3], GLfloat n2[3], GLfloat n3[3],
    GLfloat r,
    int     depth,
    GLenum  shade_type)
{
    if (--depth < 0) {
	glBegin(shade_type);
	n3fv(n3);  v3f(r * n3[0], r * n3[1], r * n3[2]);
	n3fv(n2);  v3f(r * n2[0], r * n2[1], r * n2[2]);
	n3fv(n1);  v3f(r * n1[0], r * n1[1], r * n1[2]);
	glEnd();
	return;
    }
    
    GLfloat n12[3], n23[3], n31[3];
    for (int i = 0; i < 3; i++) {
	n12[i] = n1[i] + n2[i];
	n23[i] = n2[i] + n3[i];
	n31[i] = n3[i] + n1[i];
    }
    normalize(n12);
    normalize(n23);
    normalize(n31);
    
    subdivide(n1,  n12, n31, r, depth, shade_type);
    subdivide(n2,  n23, n12, r, depth, shade_type);
    subdivide(n3,  n31, n23, r, depth, shade_type);
    subdivide(n12, n23, n31, r, depth, shade_type);
}


void
WireSphere_icosahedron(
    GLfloat r,
    int     depth)
{
    for (int i = 0; i < 20; i++) {
	subdivide(&vdata[tindices[i][0]][0],
		  &vdata[tindices[i][1]][0],
		  &vdata[tindices[i][2]][0], r, depth, GL_LINE_LOOP);
    }
}


void
SolidSphere_icosahedron(
    GLfloat r,
    int     depth)
{
    for (int i = 0; i < 20; i++) {
	subdivide(&vdata[tindices[i][0]][0],
		  &vdata[tindices[i][1]][0],
		  &vdata[tindices[i][2]][0], r, depth, GL_TRIANGLES);
    }
}

