/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

// This file contains the code based on GLUT 3.1
///* Copyright (c) Mark J. Kilgard, 1994. */

#ifndef GLEX_H
#define GLEX_H
#ifdef __GNUG__
#pragma interface
#endif

#include <stddef.h>
#include <string>
#include <GL/gl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


struct PackingMode
{
    GLint swap_bytes, lsb_first, row_length;
    GLint skip_rows, skip_pixels, alignment;
};

inline void v3f(GLfloat x, GLfloat y, GLfloat z) { glVertex3f(x, y, z); }
inline void n3f(GLfloat x, GLfloat y, GLfloat z) { glNormal3f(x, y, z); }
inline void v3fv(GLfloat v[3])       { glVertex3fv(v); }
inline void n3fv(GLfloat v[3])       { glNormal3fv(v); }
inline void Color3fv(GLfloat c[3])   { glColor3f(c[0], c[1], c[2]); }
inline void Color4fv(GLfloat c[4])   { glColor4f(c[0], c[1], c[2], c[3]); }
inline void ClearColor(GLfloat c[4]) { glClearColor(c[0], c[1], c[2], c[3]); }
inline void FColor3fv(GLfloat c[3], GLfloat f)
    { glColor3f(f * c[0], f * c[1], f * c[2]); }

Colormap GetShareableColormap(Display*, XVisualInfo*);
void GetUnpackingMode(PackingMode&);
void SetUnpackingMode(PackingMode&);

void WireSphere (GLfloat r, GLint slices = 16, GLint stacks = 16);
void SolidSphere(GLfloat r, GLint slices = 16, GLint stacks = 16);
void WireSphere_icosahedron (GLfloat r, int depth = 0);
void SolidSphere_icosahedron(GLfloat r, int depth = 0);


#define GLEX_SGLBUF  GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1, \
                     GLX_RGBA, GLX_DEPTH_SIZE, 16
#define GLEX_DBLBUF  GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1, \
                     GLX_RGBA, GLX_DEPTH_SIZE, 16, GLX_DOUBLEBUFFER

void StrokeText(GLfloat x, GLfloat y, GLfloat z, GLfloat angle, GLfloat zoom,
		string str, size_t len = 0);

//////////////////////////////////////////////////////////////////////////////
//  The following structures are taken from glut_stroke.c in GLUT 3.1
//
///* Copyright (c) Mark J. Kilgard, 1994. */
//
///* This program is freely distributable without licensing fees 
//   and is provided without guarantee or warrantee expressed or 
//   implied. This program is -not- in the public domain. */

typedef struct {
    float x;
    float y;
} CoordRec, *CoordPtr;

typedef struct {
    int num_coords;
    CoordPtr coord;
} StrokeRec, *StrokePtr;

typedef struct {
    int num_strokes;
    StrokePtr stroke;
    float center;
    float right;
} StrokeCharRec, *StrokeCharPtr;

typedef struct {
    char *name;
    int num_chars;
    StrokeCharPtr ch;
    float top;
    float bottom;
} StrokeFontRec, *StrokeFontPtr;

//////////////////////////////////////////////////////////////////////////////


void StrokeCharacter(StrokeFontPtr fontinfo, int c);

extern StrokeFontRec StrokeRoman;
extern StrokeFontRec StrokeMonoRoman;

#endif /* GLEX_H */
